#include <dshow.h>
#include <tuner.h>

#include <math.h>
#include <stdio.h>

static const GUID CLSID_KSCATEGORY_BDA_NETWORK_TUNER
 = {0x71985f48,0x1ca1,0x11d3,{0x9c,0xc8,0x00,0xc0,0x4f,0x79,0x71,0xe0}};
static const GUID CLSID_KSCATEGORY_BDA_RECEIVER_COMPONENT
  = {0xFD0A5AF4,0xB41D,0x11d2,{0x9C,0x95,0x00,0xC0,0x4F,0x79,0x71,0xE0}};
static const GUID CLSID_KSCATEGORY_BDA_TRANSPORT_INFORMATION
= {0xA2E3074F,0x6C3D,0x11d3,{0xB6,0x53,0x00,0xC0,0x4F,0x79,0x49,0x8E}};
static const GUID CLSID_PropSet
 = {0x9E1781E1,0x9CB1,0x4407,{0xBB,0xCE,0x54,0x26,0xC8,0xD0,0x0A,0x4B}};


static IBaseFilter          *pNProvider= NULL;
static IDVBTuningSpace      *pTS = NULL;
static IBaseFilter          *pBDATuner = NULL;
static IBaseFilter          *pBDACapture = NULL;
static IBaseFilter          *pM2TIF = NULL;
static IMpeg2Demultiplexer  *pM2D = NULL;


static IBaseFilter *find_dev(
  REFCLSID category,BSTR varFriendlyName)
{
  HRESULT hr;
  ICreateDevEnum *pSysDevEnum;
  IEnumMoniker *pEnumCat;
  IMoniker *pMoniker;
  IPropertyBag *pPropBag;
  IBaseFilter *pFilter=NULL;
  ULONG cFetched;
  VARIANT varName;

  hr = CoCreateInstance(
    &CLSID_SystemDeviceEnum,NULL,CLSCTX_INPROC,
    &IID_ICreateDevEnum, (void **)&pSysDevEnum);
  if (hr != S_OK) {
    fprintf(stderr,"find_dev: CoCreateInstance\n");
    return NULL;
  }

  hr = pSysDevEnum->lpVtbl->CreateClassEnumerator(pSysDevEnum,
    category,&pEnumCat,0);
  if (hr != S_OK) {
    pEnumCat->lpVtbl->Release(pEnumCat);
    fprintf(stderr,"find_dev: CreateClassEnumerator\n");
    return NULL;
  }

  VariantInit(&varName);
  while(pEnumCat->lpVtbl->Next(pEnumCat,1,&pMoniker,&cFetched) == S_OK) {

    hr = pMoniker->lpVtbl->BindToStorage(pMoniker,
      0,0,&IID_IPropertyBag,(void **)&pPropBag);
    if (hr != S_OK) {
      pMoniker->lpVtbl->Release(pMoniker);
      break;
    }
    hr = pPropBag->lpVtbl->Read(pPropBag,L"FriendlyName",&varName,0);
    pPropBag->lpVtbl->Release(pPropBag);

    if (hr != S_OK || wcscmp(varName.bstrVal,varFriendlyName) != 0) {
      pMoniker->lpVtbl->Release(pMoniker);
      continue;
    }

    hr = pMoniker->lpVtbl->BindToObject(pMoniker,
      NULL,NULL,&IID_IBaseFilter,(void**)&pFilter);
    pMoniker->lpVtbl->Release(pMoniker);
    if (hr == S_OK) break;
  }

  VariantClear(&varName);
  pEnumCat->lpVtbl->Release(pEnumCat);
  pSysDevEnum->lpVtbl->Release(pSysDevEnum);
  return pFilter;
}

static HRESULT create_filters(void)
{
  HRESULT hr;
  ITuner *pTuner;

  hr = CoCreateInstance(
    &CLSID_DVBTNetworkProvider,NULL,CLSCTX_INPROC,&IID_IBaseFilter,
    (LPVOID *)&pNProvider);
  if (hr != S_OK) {
    fprintf(stderr,"create_filters: CLSID_DVBTNetworkProvider\n");
    return hr;
  }

  hr = CoCreateInstance(
    &CLSID_DVBTuningSpace,NULL,CLSCTX_INPROC,&IID_IDVBTuningSpace,
    (LPVOID *)&pTS);
  if (hr != S_OK) {
    fprintf(stderr,"create_filters: CLSID_DVBTuningSpace\n");
    return hr;
  }
  pTS->lpVtbl->put__NetworkType(pTS,&CLSID_DVBTNetworkProvider);
  pTS->lpVtbl->put_SystemType(pTS,DVB_Terrestrial);

  hr = pNProvider->lpVtbl->QueryInterface(pNProvider,
    &IID_ITuner,(LPVOID*)&pTuner);
  if (hr != S_OK) {
    fprintf(stderr,"create_filters: IID_Tuner\n");
    return hr;
  }
  pTuner->lpVtbl->put_TuningSpace(pTuner,(ITuningSpace*)pTS);
  pTuner->lpVtbl->Release(pTuner);
  if (hr != S_OK) {
    fprintf(stderr,"create_filters: put_TuningSpace\n");
    return hr;
  }

  pBDATuner = find_dev(&CLSID_KSCATEGORY_BDA_NETWORK_TUNER,
    L"SKNET HDTV BDA Digital Tuner_0");
  if (pBDATuner == NULL) {
    fprintf(stderr,"create_filters: pBDATuner\n");
    return S_FALSE;
  }

  pBDACapture = find_dev(&CLSID_KSCATEGORY_BDA_RECEIVER_COMPONENT,
    L"SKNET HDTV BDA Digital Capture_0");
  if (pBDATuner == NULL) {
    fprintf(stderr,"create_filters: pBDACapture\n");
    return S_FALSE;
  }

  pM2TIF = find_dev(&CLSID_KSCATEGORY_BDA_TRANSPORT_INFORMATION,
    L"BDA MPEG2 Transport Information Filter");
  if (pM2TIF == NULL) {
    fprintf(stderr,"create_filters: pM2TIF\n");
    return S_FALSE;
  }

  hr = CoCreateInstance(
    &CLSID_MPEG2Demultiplexer,NULL,CLSCTX_INPROC,&IID_IBaseFilter,
    (LPVOID *)&pM2D);
  if (hr != S_OK) {
    fprintf(stderr,"create_filters: CLSID_MPEG2Demultiplexer\n");
    return hr;
  }

  return S_OK;
}

static int clear_graph(IGraphBuilder *pGB)
{
  IFilterGraph *pFG;
  IEnumFilters *pEF;
  IBaseFilter *pBF;
  HRESULT hr;

  hr = pGB->lpVtbl->QueryInterface(pGB,
    &IID_IFilterGraph,(void**)&pFG);
  if (hr != S_OK) {
    fprintf(stderr,"clear_graph: IID_IFilterGraph\n");
    return FALSE;
  }
  hr = pFG->lpVtbl->EnumFilters(pFG,&pEF);
  if (hr != S_OK) {
    pFG->lpVtbl->Release(pFG);
    fprintf(stderr,"clear_graph: EnumFilters\n");
    return FALSE;
  }
  while(pEF->lpVtbl->Next(pEF,1,&pBF,NULL) == S_OK) {
    pFG->lpVtbl->RemoveFilter(pFG,pBF);
    pBF->lpVtbl->Release(pBF);
    pEF->lpVtbl->Reset(pEF);
  }
  pEF->lpVtbl->Release(pEF);
  pFG->lpVtbl->Release(pFG);
  return TRUE;
}

static HRESULT add_filters(IGraphBuilder *pGB)
{
  HRESULT hr;

  hr = pGB->lpVtbl->AddFilter(pGB,pNProvider,L"Microsoft DVB-T Network Provider");
  if (hr != S_OK) {
    fprintf(stderr,"add_filters: pNProvider\n");
    return hr;
  }

  hr = pGB->lpVtbl->AddFilter(pGB,pBDATuner,L"SKNET HDTV BDA Digital Tuner_0");
  if (hr != S_OK) {
    fprintf(stderr,"add_filters: pBDATuner\n");
    return hr;
  }

  hr = pGB->lpVtbl->AddFilter(pGB,pBDACapture,L"SKNET HDTV BDA Digital Capture_0");
  if (hr != S_OK) {
    fprintf(stderr,"add_filters: pBDACapture\n");
    return hr;
  }

  hr = pGB->lpVtbl->AddFilter(pGB,(IBaseFilter*)pM2D,L"MPEG-2 Demultiplexer");
  if (hr != S_OK) {
    fprintf(stderr,"add_filters: pM2D\n");
    return hr;
  }

  hr = pGB->lpVtbl->AddFilter(pGB,pM2TIF,L"BDA MPEG-2 Transport Information Filter");
  if (hr != S_OK) {
    fprintf(stderr,"add_filters: pM2TIF\n");
    return hr;
  }

  return S_OK;
}

static HRESULT connect_filters_directly(IGraphBuilder *pGB,
  IBaseFilter *src,IPin *pIn)
{
  IEnumPins *ep;
  IPin *pOut;
  PIN_DIRECTION pd;
  HRESULT hr;

  hr = src->lpVtbl->EnumPins(src,&ep);
  if (hr != S_OK) {
    fprintf(stderr,"connect_filters_directly: EnumPins\n");
    return S_FALSE;
  }
  while(ep->lpVtbl->Next(ep,1,&pOut,NULL) == S_OK) {
    hr = pOut->lpVtbl->QueryDirection(pOut,&pd);
    if (hr != S_OK) {
      fprintf(stderr,"connect_filters_directly: QueryDirection\n");
      pOut->lpVtbl->Release(pOut); ep->lpVtbl->Release(ep);
      return 0;
    }
    if (pd == PINDIR_OUTPUT &&
        pGB->lpVtbl->ConnectDirect(pGB,pOut,pIn,NULL) == S_OK) {
      pOut->lpVtbl->Release(pOut); ep->lpVtbl->Release(ep);
      return S_OK;
    }
    pOut->lpVtbl->Release(pOut);
  }
  ep->lpVtbl->Release(ep);
  return S_FALSE;
}

static int connect_filters(IGraphBuilder *pGB,
  IBaseFilter *src,IBaseFilter *dest)
{
  IEnumPins *ep;
  IPin      *pp;
  PIN_DIRECTION pd;
  HRESULT hr;
  int count = 0;

  hr = dest->lpVtbl->EnumPins(dest,&ep);
  if (hr != S_OK) {
    fprintf(stderr,"connect_filters: EnumPins\n");
    return count;
  }
  while(ep->lpVtbl->Next(ep,1,&pp,NULL) == S_OK) {
    hr = pp->lpVtbl->QueryDirection(pp,&pd);
    if (hr != S_OK) {
      fprintf(stderr,"connect_filters: QueryDirection\n");
      pp->lpVtbl->Release(pp); ep->lpVtbl->Release(ep);
      return 0;
    }
    if (pd == PINDIR_INPUT &&
        connect_filters_directly(pGB,src,pp) == S_OK)
      count++;
    pp->lpVtbl->Release(pp);
  }
  ep->lpVtbl->Release(ep);
  return count;
}


static HRESULT request_tuning(ULONG freq)
{
  HRESULT hr;
  IDVBTuneRequest *pTR;
  IDVBTLocator *pLocator;
  ITuner *pTuner;

  hr = CoCreateInstance(
    &CLSID_DVBTLocator,NULL,CLSCTX_INPROC,&IID_IDVBTLocator,
    (LPVOID *)&pLocator);
  if (hr != S_OK) {
    fprintf(stderr,"request_tuning: IDVBTLocator\n");
    return hr;
  }

  hr = pTS->lpVtbl->CreateTuneRequest(pTS,
    (ITuneRequest**)&pTR);
  if (hr != S_OK) {
    pLocator->lpVtbl->Release(pLocator);
    fprintf(stderr,"request_tuning: CreateTuneRequest\n");
    return hr;
  }

  
  hr = pLocator->lpVtbl->put_CarrierFrequency(pLocator,freq);
  if (hr != S_OK) {
    fprintf(stderr,"request_tuning: put_CarrierFrequency\n");
    goto do_tune1;
  }

  hr = pTR->lpVtbl->put_Locator(pTR,
    (ILocator*)pLocator);
  if (hr != S_OK) {
    fprintf(stderr,"request_tuning: put_Locator\n");
    goto do_tune1;
  }

  hr = pNProvider->lpVtbl->QueryInterface(pNProvider,
    &IID_ITuner,(LPVOID*)&pTuner);
  if (hr != S_OK) {
    fprintf(stderr,"request_tuning: IID_Tuner\n");
    goto do_tune1;
  }
  hr = pTuner->lpVtbl->put_TuneRequest(pTuner,(ITuneRequest*)pTR);
  pTuner->lpVtbl->Release(pTuner);
  if (hr != S_OK) {
    fprintf(stderr,"request_tuning: put_TuneRequest\n");
    goto do_tune1;
  }

do_tune1:
  pTR->lpVtbl->Release(pTR);
  pLocator->lpVtbl->Release(pLocator);
  return hr;
}

static HRESULT tune(int ch,BOOL catv)
{
  ULONG freq;

  if (ch < 13 || ch > 63)
    return S_FALSE;
  if (catv) {
    if (ch <= 22 )
      freq = 33000+6000*ch;
    else
      freq = 87000+6000*ch;
  } else
    freq = 395000+6000*ch;

  return request_tuning(freq);
}

static int build(IGraphBuilder *pGB)
{
  if (create_filters() != S_OK)
    return FALSE;

  if (add_filters(pGB) != S_OK)
    return FALSE;

  if (connect_filters(pGB,pNProvider,pBDATuner) == 0) {
    return FALSE;
  }

  if (connect_filters(pGB,pBDATuner,pBDACapture) == 0) {
    return FALSE;
  }

  if (connect_filters(pGB,pBDACapture,(IBaseFilter*)pM2D) == 0) {
    return FALSE;
  }

  if (connect_filters(pGB,(IBaseFilter*)pM2D,pM2TIF) == 0) {
    return FALSE;
  }

  return TRUE;
}

static int get_level(DWORD param)
{
  HRESULT hr;
  IKsPropertySet *pIKsPropertySet;
  BYTE buf[4];
  DWORD m;

  hr = pBDATuner->lpVtbl->QueryInterface(pBDATuner,
    &IID_IKsPropertySet,(LPVOID*)&pIKsPropertySet);
  if (hr != S_OK) {
    fprintf(stderr,"get_level: QueryInterface\n");
    return -2;
  }

  hr = pIKsPropertySet->lpVtbl->Set(pIKsPropertySet,
    &CLSID_PropSet,3,NULL,0,&param,4);
  if (hr != S_OK) {
    fprintf(stderr,"get_level: IKsPropertySet.Set hr=%08X\n",hr);
    goto get_level_last;
  }

  hr = pIKsPropertySet->lpVtbl->Get(pIKsPropertySet,
    &CLSID_PropSet,3,&param,4,&buf,4,&m);
  if (hr == S_OK && m == 4) {
    pIKsPropertySet->lpVtbl->Release(pIKsPropertySet);
    return buf[0];
  }

get_level_last:
  param = 1;
  hr = pIKsPropertySet->lpVtbl->Set(pIKsPropertySet,
    &CLSID_PropSet,4,NULL,0,&param,4);
  pIKsPropertySet->lpVtbl->Release(pIKsPropertySet);
  if (hr != S_OK) {
    fprintf(stderr,"get_level: IKsPropertySet.Set (2) hr=%08X\n",hr);
  }
  return -2;
}

static int get_level1(void)
{
  int m,r;

  if (pBDATuner == NULL)
    return -2;
  m = get_level(0x8b20);
  if (m < 0)
    return m;
  r = m << 8;
  m = get_level(0x8c20);
  if (m < 0)
    return m;
  r += m;
  m = get_level(0x8d20);
  if (m < 0)
    return m;

  r = (r << 8)+m;
  if (r == 0)
    return -1;
  r = (int)floor(20*log10(5505024.0/r)+0.5);
  if (r > 99)
    return 99;
  return r;
}

static double get_level2(void)
{
  int m,r;
  double d;

  if (pBDATuner == NULL)
    return -2.0;
  m = get_level(0xa020);
  if (m < 0)
    return m;
  r = m << 8;
  m = get_level(0xa120);
  if (m < 0)
    return m;
  r += m;
  m = get_level(0xa220);
  if (m < 0)
    return m;
  d = (r << 8)+m;

  m = get_level(0xa620);
  if (m < 0)
    return m;
  r = m << 8;
  m = get_level(0xa720);
  if (m < 0)
    return m;
  return d/((r+m)*1632.0);
}

int main(int argc,char *argv[],char *envp[])
{
  HRESULT hr;
  IGraphBuilder *pGB;
  int ch,catv,dur;
  DWORD t;
  int level1;
  double level2;

  if (argc < 3) {
    fprintf(stderr,
      "hduslvlc channel duration\n"
      "  channel =13-63 c13-c63\n"
      "  duration=1-999\n"
      );
    return EXIT_SUCCESS;
  }
  catv = FALSE;
  if (sscanf(argv[1],"%d",&ch) != 1) {
    catv = TRUE;
    if (sscanf(argv[1],"c%d",&ch) != 1 && sscanf(argv[1],"C%d",&ch) != 1)
      ch = -1;
  }
  if (ch < 13 || ch > 63) {
    fprintf(stderr,"Invalid channel.\n");
    return EXIT_FAILURE;
  }
  if (sscanf(argv[2],"%d",&dur) != 1)
    dur = -1;
  if (dur < 1 || dur > 999) {
    fprintf(stderr,"Invalid duration.\n");
    return EXIT_FAILURE;
  }

  hr = CoInitialize(NULL);
  if (hr != S_OK && hr != S_FALSE) {
    fprintf(stderr,"CoInitialize\n");
    return EXIT_FAILURE;
  }
  hr = CoCreateInstance(
    &CLSID_FilterGraph,NULL,CLSCTX_INPROC_SERVER,
    &IID_IGraphBuilder,(void **)&pGB);
  if (!build(pGB))
    goto last;
  if (tune(ch,FALSE) != S_OK)
    goto last;

  t = GetTickCount()+dur*1000;
  while (t > GetTickCount()) {
    level1 = get_level1();
    if (level1 == -2)
      goto last;
    level2 = get_level2();
    if (level1 == -2.0)
      goto last;
    if (level1 == -1)
      printf(" -- |");
    else
      printf(" %2d |",level1);
    if (level2 == -1.0)
      printf(" --------- \n");
    else
      printf(" %9.2E\n",level2);

    Sleep(1000);
  }

last: 

  if (pGB != NULL) {
    clear_graph(pGB);
    pGB->lpVtbl->Release(pGB);
  }
  if (pNProvider != NULL)
    pNProvider->lpVtbl->Release(pNProvider);
  if (pBDATuner != NULL)
    pBDATuner->lpVtbl->Release(pBDATuner);
  if (pBDACapture != NULL)
    pBDACapture->lpVtbl->Release(pBDACapture);
  if (pM2TIF != NULL)
    pM2TIF->lpVtbl->Release(pM2TIF);
  if (pM2D != NULL)
    pM2D->lpVtbl->Release(pM2D);
  if (pTS != NULL)
    pTS->lpVtbl->Release(pTS);
  CoUninitialize();
  return EXIT_SUCCESS;
}
